#include "lanhep.h"

static int cls_no=1;

static Atom cls_names[128];
static List cls_membs[128];
static char cls_ini[128];
static List cls_lens=0;

void prt2cls(Atom *p)
{
	Term prop=GetAtomProperty(*p,OPR_CLASS);
	if(prop)
	{
		*p=cls_names[IntegerValue(CompoundArg1(prop))];
	}
}

int ProcClass(Term cl, Term ind)
{
	Term t=ConsumeCompoundArg(cl,1);
	Atom clnm, anti=0;
	List clmem, l1;
	int i, ms=0,noms=0,spin=-1;
	char cbuf[128];
	FreeAtomic(cl);
	
	for(i=0;i<128;i++) cls_ini[i]=0;
	
	if(is_compound(t) && CompoundArity(t)==2 && CompoundName(t)==OPR_COMMA)
		{
		Term a1,a2;
		a1=ConsumeCompoundArg(t,1);
		a2=ConsumeCompoundArg(t,2);
		FreeAtomic(t);
		ProcClass(MakeCompound1(OPR_CLASS,a1),0);
		ProcClass(MakeCompound1(OPR_CLASS,a2),0);
		return 0;
		}

	if(!is_compound(t) || CompoundArity(t)!=2 || CompoundName(t)!=OPR_EQSIGN
			|| !is_atom(CompoundArg1(t)) || !is_list(CompoundArg2(t))
			|| CompoundArg2(t)==0 )
		{
		ErrorInfo(325);
		printf("bad argument in class statement\n");
		return 0;
		}

	clnm=CompoundArg1(t);
	clmem=ConsumeCompoundArg(t,2);
	FreeAtomic(t);
	if(GetAtomProperty(clnm,PROP_TYPE))
	{
		ErrorInfo(325);
		printf("class: name %s is already in use.\n",AtomValue(clnm));
		return 0;
	}
	
	t=GetAtomProperty(ListFirst(clmem),PROP_INDEX);
	for(l1=clmem;l1;l1=ListTail(l1))
	{
		if(!is_atom(ListFirst(l1)) || !is_particle(ListFirst(l1),0))
		{
			ErrorInfo(325);
			printf("class: '");WriteTerm(ListFirst(l1));
			printf("' is not a particle.\n");
			return 0;
		}
		if(!EqualTerms(t,GetAtomProperty(ListFirst(l1),PROP_INDEX)))
		{
			ErrorInfo(325);
			printf("class %s: ",AtomValue(clnm));
			printf("'%s' and '%s' have different indices.\n",
					AtomValue(ListFirst(clmem)),AtomValue(ListFirst(l1)));
			return 0;
		}
		if(GetAtomProperty(ListFirst(l1),OPR_CLASS))
		{
			ErrorInfo(325);
			printf("class %s: ",AtomValue(clnm));
			printf("'%s' is already a member of another class.\n",
					AtomValue(ListFirst(l1)));
			return 0;
		}
			
			
	}
	t=AppendLast(CopyTerm(t),MakeCompound2(A_I,MakeCompound2(OPR_WILD,
			NewInteger(ListLength(clmem)),NewInteger(ListLength(clmem))),0));
	SetAtomProperty(clnm,PROP_INDEX,t);
	SetAtomProperty(clnm,OPR_CLASS,NewInteger(cls_no));
	if(!ListMember(cls_lens,NewInteger(ListLength(clmem))))
		cls_lens=AppendLast(cls_lens,NewInteger(ListLength(clmem)));
	SetAtomProperty(clnm,PROP_TYPE,MakeCompound1(OPR_CLASS,NewInteger(cls_no)));
	
	for(l1=clmem,i=1;l1;l1=ListTail(l1),i++)
	{
		int csp=-1;
		Term p2=GetAtomProperty(ListFirst(l1),PROP_TYPE);
		if(CompoundName(p2)==OPR_FIELD)
		{
			if(IntegerValue(CompoundArg2(p2))<4)
				csp=0;
			else csp=1;
			p2=GetAtomProperty(CompoundArg1(p2),PROP_TYPE);
		}
		else
			csp=IntegerValue(CompoundArgN(p2,4));
		if(CompoundArgN(p2,5))
			ms++;
		else
			noms++;
		if(spin==-1)
			spin=csp;
		else if(spin!=csp)
		{
			ErrorInfo(144);
			printf("class %s: different spin particles.\n",
					AtomValue(clnm));
		}
		SetAtomProperty(ListFirst(l1),OPR_CLASS,MakeCompound2(OPR_CLASS,
				NewInteger(cls_no),NewInteger(i)));
	}
	
	t=GetAtomProperty(ListFirst(clmem),A_ANTI);
	for(i=1;i<cls_no;i++)
		if(ListFirst(cls_membs[i])==t)
		{
			List l2;
			if(ListLength(clmem)!=ListLength(cls_membs[i]))
			{
				ErrorInfo(144);
				printf("classes %s/%s: different lengths.\n",
						AtomValue(cls_names[i]),AtomValue(clnm));
				return 0;
			}
			for(l1=clmem,l2=cls_membs[i];l1;l1=ListTail(l1),l2=ListTail(l2))
			{
				if(GetAtomProperty(ListFirst(l1),A_ANTI)!=ListFirst(l2))
				{
					ErrorInfo(900);
					printf("classes %s/%s: '%s'/'%s' are not conjugated.\n",
							AtomValue(cls_names[i]),AtomValue(clnm),
							AtomValue(ListFirst(l1)),AtomValue(ListFirst(l2)));
					return 0;
				}
			}
			anti=cls_names[i];
			break;
		}
	
	if(ms&&noms)
	{
		WarningInfo(0);
		printf("class %s: mixing massive and massless particles.\n",
				AtomValue(clnm));
	}
	
	if(anti)
	{
		t=GetAtomProperty(anti,PROP_TYPE);
		SetCompoundArg(t,2,clnm);
		SetAtomProperty(clnm,PROP_TYPE,t);
	}
	else
	{	
		t=MakeCompound(OPR_PARTICLE,8);
		SetCompoundArg(t,1,clnm);
		SetCompoundArg(t,2,clnm);
		sWriteTerm(cbuf,clmem);
		SetCompoundArg(t,3,NewAtom(cbuf,0));
		SetCompoundArg(t,4,NewInteger(spin));		
		if(ms)
		{
			sprintf(cbuf,"%sMass",AtomValue(clnm));
			SetCompoundArg(t,5,NewAtom(cbuf,0));
		}
		SetAtomProperty(clnm,PROP_TYPE,t);
	}
	
	t=GetAtomProperty(ListFirst(clmem),A_COLOR);
	if(t)
		SetAtomProperty(clnm,A_COLOR,t);
	
	cls_names[cls_no]=clnm;
	cls_membs[cls_no]=clmem;
	cls_no++;
	if(cls_no==128)
	{
		puts("Internal error - too many classes");
		exit(0);
	}
	return 0;	
}

void cls_wrt_nms(FILE *f, List cpml)
{
	List l1;
	int i;
	for(i=1;i<cls_no;i++)
	{
		Term fp=GetAtomProperty(cls_names[i],A_FANUM);
		int cl, no, j, m=0;
		if(fp==0)
			continue;
		cl=IntegerValue(ListFirst(fp));
		no=IntegerValue(ListFirst(ListTail(fp)));
		if(no<0) {m=1; no=-no;}
		for(l1=cls_membs[i],j=1;l1;l1=ListTail(l1),j++)
		{
			fprintf(f,"prt[\"");fWriteTerm(f,ListFirst(l1));fprintf(f,"\"] = ");
			if(m) fprintf(f,"-");
			switch(cl){
				case 0: fprintf(f,"S[%d, {%d}]\n",no,j); break;
				case 1: fprintf(f,"F[%d, {%d}]\n",no,j); break;
				case 2: fprintf(f,"V[%d, {%d}]\n",no,j); break;
					  }
		}
	}
	fprintf(f,"\n");
	
	for(i=1;i<cls_no;i++)
	{
		Term fp=GetAtomProperty(cls_names[i],PROP_TYPE);
		int j;
		Atom ms=CompoundArgN(fp,5);
		if(ms==0)
			continue;
		
		if(CompoundArg1(fp)!=CompoundArg2(fp) && CompoundArg2(fp)==cls_names[i])
			continue;
		
		for(j=1,l1=cls_membs[i];l1;l1=ListTail(l1),j++)
		{
			Term prp=GetAtomProperty(ListFirst(l1),PROP_TYPE);
			if(CompoundName(prp)==OPR_FIELD)
				prp=GetAtomProperty(CompoundArg1(prp),PROP_TYPE);
			fprintf(f,"%s[%d] = %s;\n",AtomValue(ms),j,CompoundArgN(prp,5)?
					AtomValue(CompoundArgN(prp,5)):"0");
		}
		
		for(l1=cpml;l1;l1=ListTail(l1))
		{
			if(ListFirst(l1)==ms)
			{
				fprintf(f,"%s[gen_, _] := %s[gen];\n",
						AtomValue(ms),AtomValue(ms));
				ChangeList(l1,0);
			}
		}
	}
		
	for(l1=cpml;l1;l1=ListTail(l1))
		{
			if(ListFirst(l1)==0)
				continue;
			fprintf(f,"%s[_] = %s\n",AtomValue(ListFirst(l1)),
					AtomValue(ListFirst(l1)));
		}
	fprintf(f,"\n");
}

void cls_wrt_ind(FILE *f)
{
	int l;
	for(l=cls_lens;l;l=ListTail(l))
		fprintf(f,"IndexRange[ Index[TheyAre%d] ] = Range[%d]\n",
				IntegerValue(ListFirst(l)),IntegerValue(ListFirst(l)));
}

int cls_prt_info(Term *prp, Atom *a)
{
	Term cl=GetAtomProperty(*a,OPR_CLASS);
	int  n;
	if(cl==0)
		return 0;
	n=IntegerValue(CompoundArg1(cl));
	if(cls_ini[n])
	{
		*prp=0;
		return 0;
	}
	cls_ini[n]=1;
	*prp=GetAtomProperty(cls_names[n],PROP_TYPE);
	*a=cls_names[n];
	return ListLength(cls_membs[n]);
}

extern int sp_cmp(Term,Term);

static List mtr_known=0;
static List mtr_no=0, mtr_fit=0;

static void setzprm(List ml)
{
	for(;ml;ml=ListTail(ml))
	{
		List l1;
		List l2=ConsumeCompoundArg(ListFirst(ml),2);
		List l3=ConsumeCompoundArg(ListFirst(ml),3);
		l3=SortedList(l3,sp_cmp);
		for(l1=l2;l1;l1=ListTail(l1))
			if(GetAtomProperty(CompoundArg1(ListFirst(l1)),A_INFINITESIMAL))
			{
				l3=AppendFirst(l3,ListFirst(l1));
				ChangeList(l1,0);
				l2=CutFromList(l2,l1);
				break;
			}
		for(l1=l2;l1;l1=ListTail(l1))
			if(CompoundArg1(ListFirst(l1))==A_I)
			{
				l3=AppendFirst(l3,ListFirst(l1));
				ChangeList(l1,0);
				l2=CutFromList(l2,l1);
				break;
			}
		SetCompoundArg(ListFirst(ml),2,l2);
		SetCompoundArg(ListFirst(ml),3,l3);
	}
}

static void unsetzprm(List ml)
{
	for(;ml;ml=ListTail(ml))
	{
		List l1;
		List l2=ConsumeCompoundArg(ListFirst(ml),2);
		List l3=ConsumeCompoundArg(ListFirst(ml),3);
		for(l1=l3;l1;l1=ListTail(l1))
			if(CompoundArg1(ListFirst(l1))==A_I)
			{
				l2=AppendFirst(l2,ListFirst(l1));
				ChangeList(l1,0);
				l3=CutFromList(l3,l1);
				break;
			}
		for(l1=l3;l1;l1=ListTail(l1))
			if(CompoundName(ListFirst(l1))==OPR_POW)
			{
				l2=AppendFirst(l2,ListFirst(l1));
				ChangeList(l1,0);
				l3=CutFromList(l3,l1);
				break;
			}
		SetCompoundArg(ListFirst(ml),2,l2);
		SetCompoundArg(ListFirst(ml),3,l3);
	}
}

static void add_elem(List *el, List prt, List ml)
{
	List l1;
	List pl=0;
	
	for(l1=prt;l1;l1=ListTail(l1))
	{
		Term prp=GetAtomProperty(CompoundArg1(ListFirst(l1)),OPR_CLASS);
		if(prp)
			pl=AppendLast(pl,CompoundArg2(prp));
	}
	
	for(l1=ml;l1;l1=ListTail(l1))
	{
		Term m2=ListFirst(l1);
		Term cme;
		List lt=ConsumeCompoundArg(m2,3);
		List l2;
		for(l2=(*el);l2;l2=ListTail(l2))
			if(EqualTerms(lt,CompoundArg1(ListFirst(l2))))
				break;
		if(l2==0)
		{
			cme=MakeCompound2(OPR_MINUS,lt,0);
			*el=AppendLast(*el,cme);
		}
		else
		{
			cme=ListFirst(l2);
			FreeAtomic(lt);
		}
		lt=ConsumeCompoundArg(cme,2);
		for(l2=lt;l2;l2=ListTail(l2))
			if(EqualTerms(pl,CompoundArg1(ListFirst(l2))))
				break;
		if(l2==0)
			lt=AppendLast(lt,MakeCompound2(OPR_COLON,CopyTerm(pl),MakeList1(m2)));
		else
			AppendLast(CompoundArg2(ListFirst(l2)),m2);
		SetCompoundArg(cme,2,lt);
	}
	
	FreeAtomic(pl);
	RemoveList(ml);
}

static int mtrcmp(Term t1, Term t2)
{
	List l1=CompoundArg1(t1),l2=CompoundArg1(t2);
	for(;l1;l1=ListTail(l1),l2=ListTail(l2))
	{
		int c=IntegerValue(ListFirst(l1))-IntegerValue(ListFirst(l2));
		if(c) return c;
	}
	puts("internal error: mtrcmp(cls)");
	return 0;
}

static Atom mk_mtr(List mtr, Term *nf, List *sl)
{
	Term a2=MakeCompound(A_ALG2,5);
	List ml=0;
	List l1,l2;
	mtr=SortedList(mtr,mtrcmp);
	/*DumpList(mtr);*/
	for(l1=mtr;l1;l1=ListTail(l1))
		for(l2=CompoundArg2(ListFirst(l1));l2;l2=ListTail(l2))
			ml=AppendLast(ml,ListFirst(l2));
	SetCompoundArg(a2,5,ml);
	alg2_common_n(a2);
	alg2_common_s(a2);
	*nf=ConsumeCompoundArg(a2,2);
	*sl=ConsumeCompoundArg(a2,3);
	RemoveList(ConsumeCompoundArg(a2,5));
	if(ListLength(CompoundArg1(ListFirst(mtr)))==2)
	{
		int i, dlt=1;
		for(l1=mtr,i=1;l1;l1=ListTail(l1),i++)
		{
			Term vl=CompoundArg1(ListFirst(l1));
			if(IntegerValue(ListFirst(vl))!=i || 
					ListFirst(vl)!=ListFirst(ListTail(vl)))
			{ dlt=0;break;}
			vl=CompoundArg2(ListFirst(l1));
			if(ListLength(vl)!=1) { dlt=0; break;}
			vl=ListFirst(vl);
			if(CompoundArg1(vl)!=NewInteger(1) || CompoundArg2(vl))
			{ dlt=0; break;}
		}
		if(dlt)
		{
			FreeAtomic(mtr);
			return A_DELTA;
		}
	}

	FreeAtomic(a2);

	for(l1=mtr_known;l1;l1=ListTail(l1))
		if(EqualTerms(CompoundArg2(ListFirst(l1)),mtr))
			break;
	if(l1)
	{
		FreeAtomic(mtr);
		mtr_fit++;
		return CompoundArg1(ListFirst(l1));
	}
	else
	{
		char cbuf[16];
		Atom nm;
		sprintf(cbuf,"MTR%03d",++mtr_no);
		nm=NewAtom(cbuf,0);
		mtr_known=AppendLast(mtr_known,MakeCompound2(OPR_EQSIGN,nm,mtr));
		return nm;
	}
		
/*	WriteTerm(*nf);printf(" ");WriteTerm(*sl);puts("");
	DumpList(mtr);*/
	return A_DELTA;
}
		
void setcls_1(Term a2, List a2l)
{
	int i,ci,cf=0;
	int pno, is_cls[6];
	int ind_no[6]={1000,1000,1000,1000,1000,1000};
	Atom prts[6];
	List l1,l2,elli=0,ml=0;
	
	if(CompoundArgN(a2,5)==0 || CompoundArg1(a2)==0 ||
			ListLength(CompoundArg1(a2))==1)
		return;
		
	/*WriteTerm(CompoundArg1(a2));puts("");*/
	pno=ListLength(CompoundArg1(a2));
	for(l1=CompoundArg1(a2),i=0;l1;l1=ListTail(l1),i++)
	{
		Atom p=CompoundArg1(ListFirst(l1));
		Term prp=GetAtomProperty(p,OPR_CLASS);
		if(prp)
		{
			is_cls[i]=1;
			prts[i]=cls_names[IntegerValue(CompoundArg1(prp))];
			cf++;
		}
		else
		{
			is_cls[i]=0;
			prts[i]=p;
		}
	}
	
	if(cf==0)
		return;

	alg2_decommon_s(a2);
	alg2_decommon_n(a2);

/*	WriteTerm(a2);puts("");*/
	setzprm(CompoundArgN(a2,5));
	add_elem(&elli,CompoundArg1(a2),ConsumeCompoundArg(a2,5));
	
	for(l2=a2l;l2;l2=ListTail(l2))
	{
		Term a2c=ListFirst(l2);
		if(ListLength(CompoundArg1(a2c))!=pno)
			continue;
		for(l1=CompoundArg1(a2c),i=0;l1;l1=ListTail(l1),i++)
		{
			Atom p=CompoundArg1(ListFirst(l1));
			Term prp=GetAtomProperty(p,OPR_CLASS);
			if(prp)
				p=cls_names[IntegerValue(CompoundArg1(prp))];
			if(p!=prts[i])
				break;
		}
		if(l1)
			continue;

		alg2_decommon_s(a2c);
		alg2_decommon_n(a2c);
		
		/*WriteVertex(CompoundArg1(a2c));puts("");*/
		setzprm(CompoundArgN(a2c,5));
		add_elem(&elli,CompoundArg1(a2c),ConsumeCompoundArg(a2c,5));
	}

	cf=0;
	ci=1;
	for(l1=CompoundArg1(a2),i=0;l1;l1=ListTail(l1),i++)
	{
		if(is_cls[i])
			SetCompoundArg(ListFirst(l1),1,prts[i]);
		for(l2=CompoundArg1(CompoundArg2(ListFirst(l1)));l2;l2=ListTail(l2))
		{
			ChangeList(l2,NewInteger(ci));
			ci++;
		}
		if(is_cls[i])
		{
			List ll=ConsumeCompoundArg(CompoundArg2(ListFirst(l1)),1);
			ll=AppendLast(ll,NewInteger(ci));
			SetCompoundArg(CompoundArg2(ListFirst(l1)),1,ll);
			ind_no[cf]=ci;
			ci++; cf++;
		}
	}
	
	for(l1=elli;l1;l1=ListTail(l1))
	for(l2=CompoundArg1(ListFirst(l1));l2;l2=ListTail(l2))
		if(CompoundName(ListFirst(l2))==OPR_SPECIAL || 
					CompoundName(ListFirst(l2))==A_MOMENT)
		{
			List l3;
			for(l3=CompoundArg2(ListFirst(l2));l3;l3=ListTail(l3))
			{
				int ov=IntegerValue(ListFirst(l3));
				int dn=0;
				for(i=0;i<cf;i++)
					if(ov>=ind_no[i]-i) dn++;
				if(dn)
					ChangeList(l3,NewInteger(ov+dn));
			}
		}
	
	l2=0;
	for(i=0;i<cf;i++)
		l2=AppendLast(l2,NewInteger(ind_no[i]));
		
	for(l1=elli;l1;l1=ListTail(l1))
	{
		Term m2=MakeCompound(A_MTERM,3);
		Term nc, mtr;
		List tl,sl;
		tl=ConsumeCompoundArg(ListFirst(l1),1);
		mtr=mk_mtr(ConsumeCompoundArg(ListFirst(l1),2),&nc,&sl);
		mtr=MakeCompound2(OPR_PARAMETER,mtr,CopyTerm(l2));
		tl=AppendLast(tl,mtr);
		SetCompoundArg(m2,1,nc);
		SetCompoundArg(m2,2,sl);
		SetCompoundArg(m2,3,tl);
		ml=AppendLast(ml,m2);
	}
	FreeAtomic(l2);
	unsetzprm(ml);
	SetCompoundArg(a2,5,ml);
	alg2_common_n(a2);
	alg2_common_s(a2);
/*	WriteTerm(CompoundArg1(a2));puts("");
	WriteTerm(CompoundArgN(a2,5));puts("");*/
	FreeAtomic(elli);
}

void alg2_setcls(List a2l)
{
	List l1,l2;
	if(cls_no==1)
		return;
		
	
	for(l1=a2l;l1;l1=ListTail(l1))
		for(l2=CompoundArg1(ListFirst(l1));l2;l2=ListTail(l2))
		{
			Term prp=GetAtomProperty(CompoundArg1(ListFirst(l2)),PROP_TYPE);
			if(CompoundName(prp)==OPR_FIELD && CompoundArg2(prp)==NewInteger(4))
				SetCompoundArg(ListFirst(l2),1,CompoundArg1(prp));
		}
	
	for(l1=a2l;l1;l1=ListTail(l1))
		setcls_1(ListFirst(l1),ListTail(l1));
}

